"""
RATIONALS

Base class for QQ.

AUTHORS:
- Benjamin Antieau (2024).
"""

import flint
from jacamar.rings.elements import AbstractRingElement
from jacamar.rings.rings import AbstractRing


class Rational(AbstractRingElement):
    data_class = flint.fmpq

    def __init__(self, *args):
        if len(args) == 3:
            ring, p, q = args
            AbstractRingElement.__init__(
                self,
                ring,
                self.data_class(p, q),
            )

        elif len(args) == 2:
            ring, p = args
            AbstractRingElement.__init__(
                self,
                ring,
                p,
            )

        elif len(args) == 1:
            ring = args[0]
            AbstractRingElement.__init__(
                self,
                ring,
                0,
            )

        else:
            raise TypeError(
                f"Rational() takes 1, 2, or 3 arguments, but {len(args)} were given."
            )

    def __truediv__(self, other):
        """Returns self / other with type that of self."""
        return self.__class__(self.ring, self.data / other.data)

    def __itruediv__(self, other):
        """Modifies self in place by self / other."""
        self.data /= other.data
        return self

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.__str__()


class RationalRing(AbstractRing):
    def __init__(self):
        AbstractRing.__init__(self, Rational, exact=True)
        self.zero = self(0)
        self.one = self(1)

    def __call__(self, *args):
        return Rational(self, *args)

    def __str__(self):
        return "The ring of Rational numbers (via flint.fmpq)."

    def __repr__(self):
        return self.__str__()


# Create global version of the rational ring.
QQ = RationalRing()


"""
Select functions of QQ's based on flint.fmpq attributes. (https://fredrikj.net/python-flint/fmpq.html)
"""


def p(x):
    """Returns numerator of x as ZZ."""
    return x.ring(x.data.p)


def q(x):
    """Returns denominator of x as ZZ."""
    return x.ring(x.data.q)



"""
TEST_MORPHISMS

Tests for the IdentityRingMorphism class.
"""

from jacamar.rings.integers import ZZ, ZZ_py
from jacamar.rings.morphisms import AbstractRingMorphism
from jacamar.rings.rationals import QQ


class TestIdentity:
    """Tests for the Integer class."""

    def test_identity_zz(self):
        i = ZZ.identity_morphism()
        assert i(ZZ(5)) == ZZ(5)

    def test_identity_zzpy(self):
        i = ZZ_py.identity_morphism()
        assert i(ZZ_py(5)) == ZZ_py(5)

    def test_identity_qq(self):
        i = QQ.identity_morphism()
        assert i(QQ(5)) == QQ(5)
        assert i(QQ(3, 5)) == QQ(12, 20)


class TestGeneric:
    """Tests for the AbstractRingMorphism class."""

    def test_not_implemented(self):
        """Tests that __call__ is not implemented."""
        assert AbstractRingMorphism(ZZ, ZZ)(ZZ(4)) == NotImplemented



"""
INTEGERS MODULO N

Base class for ZN.

AUTHORS:
- Joseph Grantham (2024).
"""

from flint import nmod  # pylint: disable=no-name-in-module
from jacamar.rings.elements import AbstractRingElement
from jacamar.rings.rings import AbstractRing
from jacamar.rings.integers import ZZ


class IntegerModN(AbstractRingElement):
    """Element representing an integer modulo N"""

    data_class = nmod

    def __init__(self, ring, data):
        if isinstance(data, AbstractRingElement):
            data = data.data
        if isinstance(ring.modulus, AbstractRingElement):
            data = nmod(data, ring.modulus.data)
        AbstractRingElement.__init__(
            self,
            ring,
            data,
        )

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.__str__()

    def __truediv__(self, other):
        return self.__class__(self.ring, self.data / other.data)

    def __floordiv__(self, other):
        return NotImplemented

    def __divmod__(self, other):
        return NotImplemented

    def __mod__(self, other):
        return NotImplemented

    def __abs__(self):
        raise AttributeError

    def __ge__(self, other):
        return NotImplemented

    def __gt__(self, other):
        return NotImplemented

    def __le__(self, other):
        return NotImplemented

    def __lt__(self, other):
        return NotImplemented


class IntegerModNRing(AbstractRing):
    """Ring of integers modulo N"""

    def __init__(self, modulus):
        AbstractRing.__init__(self, IntegerModN, exact=True)
        if isinstance(modulus, AbstractRingElement):
            modulus = ZZ(modulus.data)
        else:
            modulus = ZZ(modulus)
        self.modulus = modulus
        self.one = self(1)
        self.zero = self(0)

    def __str__(self):
        return "The ring of Integers modulo " + str(self.modulus) + "."

    def __repr__(self):
        return self.__str__()



"""
TEST_SERIES

Tests for the SeriesData, Series, and PowerSeriesRing classes.
"""

import pytest
import flint
from jacamar.rings.polynomials import (
    PackedMonomialData,
    SparseMonomialData,
    PolynomialData,
    Polynomial,
    PolynomialRing,
)
from jacamar.rings.series import (
    SeriesData,
    Series,
    PowerSeriesRing,
)
from jacamar.rings.integers import ZZ


class TestSeriesData:
    """Tests for the SeriesData class."""

    a = PolynomialData(
        ZZ,
        {
            SparseMonomialData((1, 1, 3, 4)): ZZ(5).data,
            SparseMonomialData((0, 2, 5, 3)): ZZ(99).data,
        },
    )
    b = PolynomialData(
        ZZ,
        {
            SparseMonomialData((1, 4, 3, 4)): ZZ(5).data,
            SparseMonomialData((0, 3, 5, 5)): ZZ(99).data,
        },
    )
    f = SeriesData(ZZ, [(5, a)], 20)
    g = SeriesData(ZZ, [(8, b)], 20)
    h = SeriesData(ZZ, [(5, a), (8, b)], 20)

    def test_printing(self):
        """Tests printing."""
        assert str(self.f) == str([(5, self.a)])
        assert repr(self.f) == str(self.f)

    def test_add(self):
        """Tests __add__."""
        assert self.f + self.g == self.h

    def test_sub(self):
        """Tests __sub__."""
        assert self.f - self.g == SeriesData(ZZ, [(5, self.a), (8, -self.b)], 20)

    def test_ne(self):
        """Tests __ne__."""
        assert not self.f + self.g != self.h
        assert self.f != self.h

    def test_mul(self):
        """Tests __mul__."""
        assert self.f * self.g == SeriesData(ZZ, [(13, self.a * self.b)], 20)
        assert self.g * self.g * self.g == SeriesData(ZZ, [], 20)

        x = PolynomialData(ZZ, {SparseMonomialData((0, 1)): ZZ(1).data})
        y = PolynomialData(ZZ, {SparseMonomialData((1, 1)): ZZ(1).data})
        w = SeriesData(
            ZZ,
            [(0, PolynomialData(ZZ, {SparseMonomialData(()): ZZ(1).data})), (1, x)],
            20,
        )
        z = SeriesData(
            ZZ,
            [(0, PolynomialData(ZZ, {SparseMonomialData(()): ZZ(1).data})), (1, y)],
            20,
        )
        f = w + z
        assert f**2 == SeriesData(
            ZZ,
            [
                (0, PolynomialData(ZZ, {SparseMonomialData(()): ZZ(4).data})),
                (1, ZZ(4).data * (x + y)),
                (2, x * x + ZZ(2).data * x * y + y * y),
            ],
            20,
        )


class TestSeries:
    """Tests for the Series class."""

    s = PowerSeriesRing(base_ring=ZZ, ngens=4, prefix="x", precision_cap=5)
    x0, x1, x2, x3 = s.gens

    def test_mul(self):
        """Tests the __mul__ and __rmul__ methods."""
        assert (self.x0 + self.x1) ** ZZ(2) == self.x0 ** ZZ(2) + ZZ(
            2
        ) * self.x0 * self.x1 + self.x1 ** ZZ(2)



"""
COMPLEXES

Base class for CC.
"""

import flint
from jacamar.rings.elements import AbstractRingElement
from jacamar.rings.rings import AbstractRing
from jacamar.rings.reals import RR


class ComplexNumber(AbstractRingElement):
    """
    Base class for complex numbers, built on AbstractRingElement.
    """

    data_class = flint.acb  # type: ignore

    def __init__(self, *args):
        if len(args) == 3:
            ring, a, b = args
            AbstractRingElement.__init__(self, ring, self.data_class(a, b))

        elif len(args) == 2:
            ring, a = args
            AbstractRingElement.__init__(self, ring, self.data_class(a))

        elif len(args) == 1:
            ring = args[0]
            AbstractRingElement.__init__(self, ring, 0)

        else:
            raise TypeError(
                f"ComplexNumber() takes 1, 2, or 3 arguments, but {len(args)} were given."
            )

    def __truediv__(self, other):
        """Returns self / other with type that of self."""
        return self.__class__(self.ring, self.data / other.data)

    def __itruediv__(self, other):
        """Modifies self in place by self / other."""
        self.data /= other.data
        return self

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.__str__()


class ComplexRing(AbstractRing):
    """
    Base class for complex rings, built on AbstractRing.
    """

    def __init__(self):
        AbstractRing.__init__(self, ComplexNumber, exact=True)
        self.zero = self(0)
        self.one = self(1)

    def __call__(self, *args):
        return ComplexNumber(self, *args)

    def __str__(self):
        return "The ring of complex numbers (via flint.acb)."

    def __repr__(self):
        return self.__str__()


# Create a global version of the complex ring.
CC = ComplexRing()


"""
Select functions of CC's based on flint.acb attributes. (https://fredrikj.net/python-flint/acb.html)
"""


def sin(x):
    """
    Sine of x.
    """
    return x.ring(x.data.sin())


def cos(x):
    """
    Cosine of x.
    """
    return x.ring(x.data.cos())


def tan(x):
    """
    Tangent of x.
    """
    return x.ring(x.data.tan())


def asin(x):
    """
    Inverse sine of x.
    """
    return x.ring(x.data.asin())


def acos(x):
    """
    Inverse cosine of x.
    """
    return x.ring(x.data.acos())


def atan(x):
    """
    Inverse tanget of x.
    """
    return x.ring(x.data.atan())


def csc(x):
    """
    Cosecant of x.
    """
    return x.ring(x.data.csc())


def sec(x):
    """
    Secant of x.
    """
    return x.ring(x.data.sec())


def cot(x):
    """
    Cotanget of x.
    """
    return x.ring(x.data.cot())


def acsc(x):
    """
    Inverse cosecant of x.
    """
    return x.ring(x.data.acsc())


def asec(x):
    """
    Inverse secant of x.
    """
    return x.ring(x.data.asec())


def acot(x):
    """
    Inverse cotanget of x.
    """
    return x.ring(x.data.acot())


def abs_lower(x):
    """
    The absolute value of the lower bound of x.
    """
    return x.ring(x.data.abs_lower())


def abs_upper(x):
    """
    The absolute value of the upper bound of x.
    """
    return x.ring(x.data.abs_upper())


def exp(x):
    """
    The constat e raised to the power of x.
    """
    return x.ring(x.data.exp())


def log(x):
    """
    Natural logarithm of x.
    """
    return x.ring(x.data.log())


def root(x, n):
    """nth root of x."""
    return x.ring(x.data.root(n))


def sqrt(x):
    """
    Square root of x.
    """
    return x.ring(x.data.sqrt())


def real(x):
    """
    Returns the real component of x.
    """
    return RR(str(x.ring(x.data.real)))


def imag(x):
    """
    Returns the imaginary component of x.
    """
    return RR(str(x.ring(x.data.imag)))



"""
TEST_POLYNOMIALS

Tests for the MonomialData, PolynomialData, Polynomial, and PolynomialRing classes.
"""

import pytest
import timeit
import flint
from jacamar.rings.polynomials import (
    MonomialData,
    PackedMonomialData,
    SparseMonomialData,
    PolynomialData,
    Polynomial,
    PolynomialRing,
    PolynomialRingMorphism,
    SpecialPolynomial,
)
from jacamar.rings.integers import ZZ, ZZ_py
from jacamar.rings.morphisms import AbstractRingMorphism
from jacamar.rings.rationals import QQ
from jacamar.rings.reals import RR


class TestMonomialData:
    """Tests for the MonomialData class."""

    def test_from_packed_integer(self):
        """Tests that from_packed_int is not implemented."""
        assert MonomialData.from_packed_integer(5) == NotImplemented

    def test_from_sparse_tuple(self):
        """Tests that from_sparse_tuple is not implemented."""
        assert MonomialData.from_sparse_tuple((1, 2, 4, 7)) == NotImplemented

    def test_from_tuple(self):
        """Tests that from_tuple is not implemented."""
        assert MonomialData.from_tuple((0, 2, 0, 0, 7)) == NotImplemented


class TestSparseMonomial:
    """Tests for the SparseMonomialData class."""

    def test_empty(self):
        """Edge tests for the empty monomial."""
        assert SparseMonomialData(()) == SparseMonomialData(())
        m = SparseMonomialData((1, 2, 4, 7))
        assert m * SparseMonomialData(()) == m

    def test_eq(self):
        """Tests the __eq__ method."""
        assert SparseMonomialData((1, 2, 4, 7)) == SparseMonomialData((1, 2, 4, 7))
        assert SparseMonomialData((1, 2, 4, 7)) != SparseMonomialData((1, 2, 4, 8))

    def test_mul(self):
        """Tests the __mul__ method."""
        m = SparseMonomialData((1, 2, 4, 7))
        n = SparseMonomialData((0, 3, 3, 3, 4, 1))
        j = SparseMonomialData((1, 1, 1, 2, 2, 4, 2, 5))

        assert n * m == SparseMonomialData((0, 3, 1, 2, 3, 3, 4, 8))
        assert m * SparseMonomialData(()) == m

    def test_hash(self):
        """Tests the __hash__ method."""
        m = SparseMonomialData((1, 2, 4, 7))
        assert hash(m) == hash((1, 2, 4, 7))

    def test_str_and_repr(self):
        """Tests the __str__ method."""
        m = SparseMonomialData((1, 2, 4, 7))
        assert str(m) == str((1, 2, 4, 7))
        assert repr(m) == str(m)

    def test_class_methods(self):
        """Tests the class method constructors."""
        assert SparseMonomialData((1, 2, 4, 7)) == SparseMonomialData.from_sparse_tuple(
            (1, 2, 4, 7)
        )
        assert SparseMonomialData((1, 2, 4, 7)) == SparseMonomialData.from_tuple(
            (0, 2, 0, 0, 7)
        )
        assert SparseMonomialData(
            (1, 2, 4, 7)
        ) == SparseMonomialData.from_packed_integer(129127208515966992384)


class TestPackedMonomial:
    """Tests for the PackedMonomialData class."""

    def test_empty(self):
        """Edge tests for the empty monomial."""
        assert PackedMonomialData(0) == PackedMonomialData(0)
        m = PackedMonomialData.from_sparse_tuple((1, 2, 4, 7))
        assert m * PackedMonomialData(0) == m

    def test_eq(self):
        """Tests the __eq__ method."""
        assert PackedMonomialData.from_sparse_tuple(
            (1, 2, 4, 7)
        ) == PackedMonomialData.from_sparse_tuple((1, 2, 4, 7))
        assert PackedMonomialData.from_sparse_tuple(
            (1, 2, 4, 7)
        ) != PackedMonomialData.from_sparse_tuple((1, 2, 4, 8))

    def test_mul(self):
        """Tests the __mul__ method."""
        m = PackedMonomialData.from_sparse_tuple((1, 2, 4, 7))
        n = PackedMonomialData.from_sparse_tuple((0, 3, 3, 3, 4, 1))

        assert m * n == PackedMonomialData.from_sparse_tuple((0, 3, 1, 2, 3, 3, 4, 8))

    def test_hash(self):
        """Tests the __hash__ method."""
        m = PackedMonomialData.from_sparse_tuple((1, 2, 4, 7))
        # This could fail if the constant PACKING_BOUND from
        # jacamar.rings.polynomials is changed.
        assert hash(m) == hash(129127208515966992384)

    def test_str_and_repr(self):
        """Tests the __str__ method."""
        m = PackedMonomialData.from_sparse_tuple((1, 2, 4, 7))
        # This could fail if the constant PACKING_BOUND from
        # jacamar.rings.polynomials is changed.
        assert str(m) == "129127208515966992384"
        assert repr(m) == str(m)

    def test_class_methods(self):
        """Tests the class method constructors."""
        assert PackedMonomialData(
            129127208515966992384
        ) == PackedMonomialData.from_sparse_tuple((1, 2, 4, 7))
        assert PackedMonomialData(
            129127208515966992384
        ) == PackedMonomialData.from_tuple((0, 2, 0, 0, 7))
        assert PackedMonomialData(
            129127208515966992384
        ) == PackedMonomialData.from_packed_integer(129127208515966992384)


class TestPolynomialDataClass:
    """Tests for the PolynomialData class."""

    d = {
        SparseMonomialData((0, 1)): flint.fmpz(1),
        SparseMonomialData((1, 1)): flint.fmpz(1),
    }
    p = PolynomialData(ZZ, d)

    def test_mul(self):
        """Tests for the __mul__ method."""
        # (x+y)^2 == x^2 + 2xy + y^2

        assert self.p * self.p == PolynomialData(
            ZZ,
            {
                SparseMonomialData((0, 2)): flint.fmpz(1),
                SparseMonomialData((0, 1, 1, 1)): flint.fmpz(2),
                SparseMonomialData((1, 2)): flint.fmpz(1),
            },
        )

    def test_is_zero(self):
        """Tests the is_zero method."""
        assert not self.p.is_zero()
        assert PolynomialData(ZZ, {}).is_zero()
        assert PolynomialData(ZZ, {SparseMonomialData(0, 1): ZZ.zero.data}).is_zero()
        q = PolynomialData(
            ZZ,
            {
                PackedMonomialData(23512): flint.fmpz(0),
                PackedMonomialData(515167141362451626134541345): flint.fmpz(0),
            },
        )
        assert q.is_zero()
        assert not PolynomialData(ZZ, {PackedMonomialData(5): flint.fmpz(5)}).is_zero()

    def test_term_data(self):
        """Tests the term_data method."""
        assert self.d.items() == self.p.term_data()

    def test_printing(self):
        """Tests the printing methods."""
        assert str(self.p) == "{(0, 1): 1, (1, 1): 1}"
        assert repr(self.p) == str(self.p)


class TestPolynomialRing:
    """Tests for the PolynomialRing class."""

    r = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x", packed=False).to_generic()
    x0 = r.gens[0]
    x1 = r.gens[1]
    x2 = r.gens[2]
    a = r(-5) + x0 + x1 + x2

    def test_special_poly(self):
        """Tests construction of QQ and ZZ special polynomials"""
        s = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x")
        x0 = s.gens[0]
        x1 = s.gens[1]
        x2 = s.gens[2]

        a = s(3) + x0 + x1 + x2
        q = PolynomialRing(base_ring=QQ, ngens=3, prefix="x")
        x0 = s.gens[0]
        x1 = s.gens[1]
        x2 = s.gens[2]

    def test_gens_data(self):
        """Tests the correct creation of the generators."""
        assert self.r.gens[0].data == PolynomialData(
            ZZ, {SparseMonomialData((0, 1)): flint.fmpz(1)}
        )

    def test_square(self):
        """Tests that (x0+x1)**2 == x0**2 + 2*x0x1 + x1**2."""
        assert (self.x0 + self.x1) ** ZZ(2) == (
            self.x0 ** ZZ(2) + ZZ(2) * self.x0 * self.x1 + self.x1 ** ZZ(2)
        )

    def test_constructor_from_int(self):
        """Tests r(5)."""
        r = PolynomialRing(
            base_ring=ZZ,
            ngens=1,
            prefix="x",
        )

        ctx = flint.fmpq_mpoly_ctx.get(["x", "y", "z"], flint.Ordering.lex)
        m = flint.fmpq_mpoly({(1, 0, 1): 2, (1, 1, 2): 3, (0, 1, 3): 1}, ctx)
        assert str(self.r(5)) == "5"

    def test_constructor_from_base_ring(self):
        """Tests r(ZZ(5))."""

        assert str(self.r(ZZ(5))) == "5"

    def test_constructor_from_element(self):
        """Tests constructor from element."""
        assert self.r(self.a) == self.a

    def test_constructor_from_element_data(self):
        """Tests constructor from element data."""
        assert self.r(self.a.data) == self.a

    def test_constructor_from_dict(self):
        """Tests dictionary constructor."""
        f = self.r({(1, 1, 2, 1): ZZ(2), (0, 4): ZZ(9)})
        assert f == ZZ(2) * self.x1 * self.x2 + ZZ(9) * self.x0 ** ZZ(4)

    def test_packed_sparse_str(self):
        """Tests that sparse and packed versions print the same."""
        m = str(self.a * self.a)
        s = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x", packed=True).to_generic()
        x0, x1, x2 = s.gens
        b = s(-5) + x0 + x1 + x2
        assert m == str(b * b)

    def test_sub(self):
        """Tests the sub method."""
        assert self.a - self.a == self.r.zero

    def test_negative_power(self):
        """Tests that powering fails with negative input."""
        with pytest.raises(TypeError):
            self.a ** ZZ(-5)

    def test_zero_power(self):
        """Tests that powering by zero returns one."""
        assert self.a ** ZZ(0) == self.r.one

    def test_power_failure(self):
        """Tests that powering by rational or polynomial fails."""
        with pytest.raises(TypeError):
            self.a ** QQ(1, 2)
        with pytest.raises(TypeError):
            self.a**self.a


class TestSpecialPolynomialRing:
    """Tests for Special Polynomials."""

    s = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x")
    x0 = s.gens[0]
    x1 = s.gens[1]
    x2 = s.gens[2]
    a = s(3) + x0 + x1 + x2
    b = s(1) + x0 ** ZZ(3) * x2 + x1
    u = PolynomialRing(base_ring=ZZ, ngens=5, prefix="y")
    y0 = u.gens[0]
    y1 = u.gens[1]
    y2 = u.gens[2]
    y3 = u.gens[3]
    y4 = u.gens[4]

    v = y0 + ZZ(4) * y1 ** ZZ(2) + y2 + y3 + y4

    def test_eval(self):
        """Tests evaluation at a point."""
        assert self.b(ZZ(1), ZZ(1), ZZ(1)) == ZZ(3)

    def test_special_poly(self):
        """Tests normal construction."""
        assert self.a

    def test_gens_data(self):
        """Tests the correct creation of the generators."""
        assert self.s.gens[0].data == flint.fmpz_mpoly(
            {(1, 0, 0): flint.fmpz(1)}, self.s.ctx
        )
        assert self.s.gens[1].data == flint.fmpz_mpoly(
            {(0, 1, 0): flint.fmpz(1)}, self.s.ctx
        )
        assert self.s.gens[2].data == flint.fmpz_mpoly(
            {(0, 0, 1): flint.fmpz(1)}, self.s.ctx
        )

    def test_square(self):
        """Tests that (x0+x1)**2 == x0**2 + 2*x0x1 + x1**2."""
        s = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x")

        assert (self.x0 + self.x1) ** ZZ(2) == (
            self.x0 ** ZZ(2) + ZZ(2) * self.x0 * self.x1 + self.x1 ** ZZ(2)
        )

    def test_constructor_from_int(self):
        """Tests r(5)."""
        assert str(self.s(5)) == "5"

    def test_constructor_from_base_ring(self):
        """Tests r(ZZ(5))."""
        assert str(self.s(ZZ(5))) == "5"

    def test_constructor_from_element(self):
        """Tests constructor from a SpecialPolynomial element."""
        s = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x")
        assert self.s(self.a) == self.a

    def test_constructor_from_element_data(self):
        """Tests constructor from a SpecialPolynomial element data (fmpz_mpoly)."""
        s = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x")

        assert self.s(self.a.data) == self.a

    def test_packed_sparse_str(self):
        """Tests that sparse and packed versions print the same."""
        m = str(self.a * self.a)
        s = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x", packed=True)
        x0, x1, x2 = s.gens
        b = s(-5) + x0 + x1 + x2
        m == str(b * b)

    def test_sub(self):
        """Tests the sub method."""
        s = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x")

        assert self.a - self.a == self.s.zero

    # Currently, the following test leads to Python crashing.
    '''
    def test_negative_power(self):
        """Tests that powering fails with negative input."""
        with pytest.raises(ValueError):
            self.a ** ZZ(-5)
    '''

    def test_context_error(self):
        """Tests that incompatible context operations fail."""
        with pytest.raises(ValueError):
            self.a * self.v

    def test_zero_power(self):
        """Tests that powering by zero returns one."""
        assert self.a ** ZZ(0) == self.s.one

    def test_power_failure(self):
        """Tests that powering by rational or polynomial fails."""
        with pytest.raises(TypeError):
            self.a ** QQ(1, 2)
        with pytest.raises(TypeError):
            self.a**self.a

    def test_to_generic(self):
        """Tests the to_generic() method."""
        assert self.s.to_generic()


class TestLayers:
    """Tests polynomial rings over polynomial rings."""

    r = PolynomialRing(base_ring=ZZ, ngens=2, prefix="x", packed=True).to_generic()
    s = PolynomialRing(base_ring=r, ngens=2, prefix="y", packed=True)
    x0, x1 = r.gens
    y0, y1 = s.gens
    a = r(1) + x0 + x1
    b = s(1) + y0 + y1

    def test_evaluation(self):
        """Tests evaluation."""
        assert self.s(self.a)(0, 0) == self.a

    def test_multiplication(self):
        """Tests multiplication of polynomials."""
        with pytest.raises(TypeError):
            self.b * self.a

        assert self.a * self.b == self.s(self.a) + self.a * self.y0 + self.a * self.y1
        assert self.s(self.a) * self.b == self.a * self.b


class TestCalls:
    """Test evaluation of polynomials."""

    s = PolynomialRing(base_ring=ZZ, ngens=4, prefix="x", packed=False).to_generic()
    x0, x1, x2, x3 = s.gens
    f = x0 + ZZ(2) * x1 * x2 + x3

    def test_evaluation_at_coorindates(self):
        """Tests evaluation at tuples of base_ring elements."""
        assert self.f(ZZ(1), ZZ(2), ZZ(3), ZZ(4)) == ZZ(17)

    def test_evaluation_at_monomial(self):
        """Tests evaluation at a SparseMonomialData instance."""

        assert self.f(SparseMonomialData((1, 1, 2, 1))) == ZZ(2)
        assert self.f(SparseMonomialData(1, 1, 2, 1)) == ZZ(2)

    def test_evaluation_at_implicit_monomial(self):
        """
        Tests evaluation at something that can be coerced into a
        SparseMonomialData instance.
        """
        assert self.f(1, 1, 2, 1) == ZZ(0)

    def test_evaluation_at_other(self):
        """
        Tests that evaluating at something else is NotImplemented.
        """
        assert self.f(QQ(1, 2)) == NotImplemented
        assert self.f(1, 1, 2, 1, 1, 2) == NotImplemented


class TestPrinting:
    """Test printing in various situations."""

    r = PolynomialRing(base_ring=ZZ, ngens=2, prefix="x").to_generic()
    x0, x1 = r.gens
    f = r(-5) * x0 - ZZ(9) * x0 ** ZZ(5) * x1 ** ZZ(2)

    s = PolynomialRing(base_ring=r, ngens=2, prefix="y").to_generic()
    y0, y1 = s.gens
    g = s(x0 + x1) * (y0 + y1)

    h = s(x0) * (y0 + y1)

    def test_integer_coefficients(self):
        """Tests printing with integer coefficients."""
        assert str(self.f) == "- 5*x0^1 - 9*x0^5*x1^2"
        assert repr(self.f) == str(self.f)

    def test_polynomial_coefficients(self):
        """Tests printing with polynomial coefficients."""
        assert str(self.g) == "(1*x0^1 + 1*x1^1)*y0^1 + (1*x0^1 + 1*x1^1)*y1^1"

    def test_simple_polynomial_coefficients(self):
        """Tests printing with simple polynomial coefficients."""
        assert str(self.h) == "1*x0^1*y0^1 + 1*x0^1*y1^1"

    def test_printing_ring(self):
        """Tests the printing of polynomial rings."""
        assert (
            str(self.r)
            == "Ring of polynomials in 2 variables ['x0', 'x1'] with weights [1, 1] over The ring of Integers (via flint.fmpz)."
        )
        assert (
            str(self.s)
            == "Ring of polynomials in 2 variables ['y0', 'y1'] with weights [1, 1] over Ring of polynomials in 2 variables ['x0', 'x1'] with weights [1, 1] over The ring of Integers (via flint.fmpz)."
        )
        assert repr(self.r) == str(self.r)

    def test_printing_zero(self):
        """Tests printing zero."""
        assert str(self.r.zero) == "0"


class TestWeights:
    """Tests weights."""

    r = PolynomialRing(base_ring=ZZ, ngens=2, prefix="x", weights=[2, 3])

    def test_weights(self):
        """Tests that weights are correctly initialized."""
        assert self.r.weights == [2, 3]

    def test_bad_weights(self):
        """
        Tests that a ValueError is raised if the number of weights is not ngens.
        """
        with pytest.raises(ValueError):
            s = PolynomialRing(base_ring=ZZ, ngens=2, prefix="x", weights=[1, 2, 3, 4])


class TestMorphism:
    """Tests PolynomialRingMorphisms."""

    r = PolynomialRing(base_ring=ZZ, ngens=2, prefix="x").to_generic()
    x0, x1 = r.gens

    s = PolynomialRing(base_ring=ZZ, ngens=2, prefix="y").to_generic()
    y0, y1 = s.gens

    f0 = ZZ(2) * y0 ** ZZ(3) + y0 * y1
    f1 = y0 + y1

    f = PolynomialRingMorphism(
        domain=r,
        codomain=s,
        coefficient_morphism=ZZ.identity_morphism(),
        action_on_generators=[f0, f1],
    )

    def test_morphism(self):
        """Tests evaluation of morphisms."""
        assert self.f(self.x0) == self.f0
        assert self.f(self.x1) == self.f1
        assert (
            self.f(PackedMonomialData.from_sparse_tuple((0, 1, 1, 1)))
            == self.f0 * self.f1
        )

    def test_base_ring_check(self):
        """Tests the detection of incompatible base ring morphism."""
        t = PolynomialRing(base_ring=QQ, ngens=2, prefix="z")
        z0, z1 = t.gens
        with pytest.raises(TypeError):
            PolynomialRingMorphism(
                domain=self.r,
                codomain=t,
                coefficient_morphism=ZZ.identity_morphism(),
                action_on_generators=[z0, z1],
            )

    def test_bad_ngens(self):
        """Tests the detection of the wrong number of generators."""
        with pytest.raises(TypeError):
            PolynomialRingMorphism(
                domain=self.r,
                codomain=self.s,
                coefficient_morphism=ZZ.identity_morphism(),
                action_on_generators=[self.y0, self.y1, self.f0],
            )

    def test_bad_action(self):
        """Tests ring detection."""
        with pytest.raises(TypeError):
            PolynomialRingMorphism(
                domain=self.r,
                codomain=self.s,
                coefficient_morphism=ZZ.identity_morphism(),
                action_on_generators=[self.x0, self.x1],
            )

    def test_printing(self):
        """Tests printing."""
        assert repr(self.f) == str(self.f)
        assert (
            str(self.f)
            == "Homomorphism from Ring of polynomials in 2 variables ['x0', 'x1'] with weights [1, 1] over The ring of Integers (via flint.fmpz). to Ring of polynomials in 2 variables ['y0', 'y1'] with weights [1, 1] over The ring of Integers (via flint.fmpz). defined by [2*y0^3 + 1*y0^1*y1^1, 1*y0^1 + 1*y1^1] on generators."
        )


class TestCoercion:
    """Tests a map which goes from ZZ_py to QQ on coefficients."""

    r = PolynomialRing(
        base_ring=ZZ_py,
        ngens=1,
        prefix="x",
    )
    x = r.gens[0]
    s = PolynomialRing(
        base_ring=QQ,
        ngens=1,
        prefix="y",
    ).to_generic()
    y = s.gens[0]

    class ZZpyQQMorphism(AbstractRingMorphism):
        def __init__(self):
            AbstractRingMorphism.__init__(self, ZZ_py, QQ)

        def __call__(self, x):
            return self.codomain(x.data)

    fc = ZZpyQQMorphism()
    f0 = QQ(1, 3) * y ** ZZ(4) + QQ(2, 11) * y ** ZZ(9)

    f = PolynomialRingMorphism(
        domain=r,
        codomain=s,
        coefficient_morphism=fc,
        action_on_generators=[f0],
    )

    def test_call(self):
        assert self.f(self.x) == self.f0



"""
MORPHISMS

Base classes for morphisms of commutative rings.

AUTHORS:
- Benjamin Antieau (2024).
"""


class AbstractRingMorphism:
    def __init__(self, domain, codomain):
        self.domain = domain
        self.codomain = codomain

    def __call__(self, x):
        return NotImplemented


class IdentityRingMorphism:
    def __init__(self, ring):
        AbstractRingMorphism.__init__(self, ring, ring)

    def __call__(self, x):
        return x



"""
TEST_RATIONALS

Tests for the Rational and RationalRing classes.
"""

import pytest
from jacamar.rings.integers import ZZ
from jacamar.rings.rationals import QQ, p, q


class TestRational:
    """Tests for the Rational class."""

    def test_no_args(self):
        """Tests QQ()."""
        assert QQ() == QQ(0)

    def test_bad_args(self):
        """Tests that everything fails with too many arguments."""
        with pytest.raises(TypeError):
            QQ(1, 2, 3)

    def test_truediv(self):
        """Tests that x / y."""
        x = QQ(7)
        y = QQ(11)
        q = x / y
        assert str(q) == "7/11"

    def test_itruediv_notimplemented(self):
        """Tests x /= y."""
        x = QQ(7)
        y = QQ(11)
        x /= y
        assert str(x) == "7/11"

    def test_str(self):
        """Tests __str__."""
        assert str(QQ(7, 11)) == "7/11"
        assert str(QQ) == "The ring of Rational numbers (via flint.fmpq)."

    def test_repr(self):
        """Tests __repr__."""
        assert QQ(7, 11).__repr__() == "7/11"
        assert QQ.__repr__() == "The ring of Rational numbers (via flint.fmpq)."

    def test_add_with_integer(self):
        """Tests __add__ with an integer."""
        assert ZZ(1) + QQ(1, 2) == QQ(3, 2)
        with pytest.raises(TypeError):
            QQ(1, 2) + ZZ(1)


class TestRationalFunction:
    """Tests for functions in the Rational class."""

    def test_p(self):
        """Tests the p global function."""
        assert p(QQ(7, 3)) == ZZ(7)

    def test_q(self):
        """Tests the q global function."""
        assert q(QQ(7, 3)) == ZZ(3)



"""
# JACAMAR

A tiny Python computer algebra system.

JACAMAR provides three basic sets of functionality:
- a high-level interface to FLINT, via the PythonFLINT project;
- generic code for matrices;
- generic code for multivariable Polynomials, LaurentPolynomials, and PowerSeries.
Where possible, we follow the conventions of SAGE for the creation of objects.



# Philosophy

JACAMAR favors explicit, user-imposed coercion when it makes sense. In
particular, Python `int` instances are not mathematical objects and must typically be
coerced mannualy into JACAMAR objects.

Several base commutative rings are provided. At the moment, these are ZZ, ZZ_py, and QQ.
The rings ZZ and QQ wrap FLINT classes `fmpz` and `fmpq`, while ZZ_py wraps Python's `int` class.
We will be adding RR and CC in the coming days and then later integers modulo
N, finite fields, and q-adics.

New commutative rings are constructed by building PolynomialRings, LaurentPolynomialRings,
or PowerSeries rings. These rings each refer to a class of element objects. For
example, `Polynomial` is the generic element class of a PolynomialRing. The class
`Polynomial` itself is a wrapper for a lower-level _Polynomial class.

However, for those rings provided by FLINT, the data class will be the
corresponding FLINT polynomial class.

The goal is to provide a very thin, familiar interface which the user can
easily abstract and customize for their own purposes.
"""

from jacamar.matrices.matrices import *

from jacamar.rings.elements import *
from jacamar.rings.rings import *
from jacamar.rings.morphisms import *
from jacamar.rings.integers import *
from jacamar.rings.intmod import *
from jacamar.rings.polynomials import *
from jacamar.rings.rationals import *
from jacamar.rings.series import *
from jacamar.rings.reals import *
from jacamar.rings.complexes import *

__version__ = "0.1.1"



"""
TEST_COMPLEXES

Tests for the RealNumber and RealRing classes.
"""

import pytest
import flint
from jacamar.rings.complexes import (
    CC,
    sin,
    cos,
    tan,
    asin,
    imag,
    real,
    acos,
    atan,
    csc,
    sec,
    cot,
    abs_lower,
    abs_upper,
    exp,
    log,
    root,
    sqrt,
)
from jacamar.rings.reals import RR, RR_py
from jacamar.rings.integers import ZZ, ZZ_py
from jacamar.rings.polynomials import (
    SparseMonomialData,
    PolynomialData,
    Polynomial,
    PolynomialRing,
)


class TestComplexNumber:
    """Tests for the ComplexNumber class."""

    def test_no_args(self):
        """Tests CC()."""
        assert CC() == CC(0)

    def test_truediv(self):
        """Tests that x / y."""
        x = CC(1)
        y = CC(3)
        q = x / y
        assert str(q) == "[0.333333333333333 +/- 3.71e-16]"

    def test_itruediv_notimplemented(self):
        """Tests x /= y."""
        x = CC(1)
        y = CC(3)
        x /= y
        assert str(x) == "[0.333333333333333 +/- 3.71e-16]"

    def test_str(self):
        """Tests __str__."""
        assert str(CC(2 + 3j)) == "2.00000000000000 + 3.00000000000000j"
        assert str(CC) == "The ring of complex numbers (via flint.acb)."

    def test_repr(self):
        """Tests __repr__."""
        assert CC.__repr__() == "The ring of complex numbers (via flint.acb)."

    def test_mult(self):
        """Test x * y."""
        assert CC(4 + 2j) * CC(2) == CC(8 + 4j)

    def test_add(self):
        """Test x + y."""
        assert CC(3 - 2j) + CC(1 + 1j) == CC(4 - 1j)

    def test_sub(self):
        """Test x - y."""
        assert CC(5 - 2j) - CC(2 + 3j) == CC(3 - 5j)

    def test_mult_with_polynomial(self):
        """Tests __mult__ with a polynomial."""
        poly = PolynomialData(
            CC,
            {
                SparseMonomialData((0, 1)): flint.fmpz(1),
                SparseMonomialData((1, 1)): flint.fmpz(1),
            },
        )

    def test_add_with_integer(self):
        """Tests __add__ with an integer."""
        assert ZZ(1) + CC(6.765) == CC(7.765)
        with pytest.raises(TypeError):
            CC(6.765) + ZZ(1)

    def test_mult_with_integer(self):
        """Tests __mult__ with an integer."""
        assert ZZ(1) * CC(6.765) == CC(6.765)
        with pytest.raises(TypeError):
            CC(6.765) * ZZ(1)

    def test_add_with_real(self):
        """Tests __add__ with an integer."""
        assert RR(1.5) + CC(2 + 1j) == CC(3.5 + 1j)
        with pytest.raises(TypeError):
            CC(2 + 1j) + RR(1.5)

    def test_mult_with_real(self):
        """Tests __mult__ with an integer."""
        assert RR(2) * CC(1 + 1j) == CC(2 + 2j)
        with pytest.raises(TypeError):
            CC(1 + 1j) + RR(2)


class TestComplexNumberFunctions:
    """Tests flint CC functions."""

    def test_sin(self):
        """Tests the sin global function."""
        assert str(sin(CC(1))) == "[0.841470984807897 +/- 6.08e-16]"

    def test_cos(self):
        """Tests the cos global function."""
        assert str(cos(CC(1))) == str(flint.acb(1).cos())

    def test_tan(self):
        """Tests the tan function."""
        assert str(tan(CC(1))) == str(flint.acb(1).tan())

    def test_asin(self):
        """Tests the asin global function."""
        assert str(asin(CC(1))) == str(flint.acb(1).asin())

    def test_acos(self):
        """Tests the acos function."""
        assert str(acos(CC(1))) == str(flint.acb(1).acos())

    def test_atan(self):
        """Tests the atan function."""
        assert str(atan(CC(1))) == str(flint.acb(1).atan())

    def test_csc(self):
        """Tests the csc global function."""
        assert str(csc(CC(1))) == str(flint.acb(1).csc())

    def test_sec(self):
        """Tests the sec function."""
        assert str(sec(CC(1))) == str(flint.acb(1).sec())

    def test_cot(self):
        """Tests the cot function."""
        assert str(cot(CC(1))) == str(flint.acb(1).cot())

    def test_abs_upper(self):
        """Tests the abs_upper global function."""
        assert str(abs_upper(CC("1.5+/-1"))) == str(flint.acb("1.5+/-1").abs_upper())

    def test_abs_lower(self):
        """Tests the abs_lower function."""
        assert str(abs_lower(CC("1.5+/-1"))) == str(flint.acb("1.5+/-1").abs_lower())

    def test_exp(self):
        """Tests the exp global function."""
        assert str(exp(CC(2 + 1j))) == str(flint.acb(2 + 1j).exp())

    def test_log(self):
        """Tests the log function."""
        assert str(log(CC(3.3))) == str(flint.acb(3.3).log())

    def test_root(self):
        """Tests the root function."""
        assert str(root(CC(7 + 2j), 3)) == str(flint.acb(7 + 2j).root(3))

    def test_sqrt(self):
        """Tests the sqrt function."""
        assert str(sqrt(CC(4))) == str(flint.acb(4).sqrt())

    def test_imag(self):
        """Tests the sqrt function."""
        assert str(imag(CC(4 + 3j))) == str(flint.acb(4 + 3j).imag)
        assert type(imag(CC(4 + 3j))) == type(RR(3))

    def test_real(self):
        """Tests the sqrt function."""
        assert str(real(CC(4 + 3j))) == str(flint.acb(4 + 3j).real)
        assert type(real(CC(4 + 3j))) == type(RR(4))
